/*
 * Copyright (C) 2011 Google, Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY GOOGLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef ContentSecurityPolicy_h
#define ContentSecurityPolicy_h

#include "ContentSecurityPolicyResponseHeaders.h"
#include "ContentSecurityPolicyHash.h"
#include "SecurityOrigin.h"
#include "SecurityOriginData.h"
#include "SecurityOriginHash.h"
#include <wtf/HashSet.h>
#include <wtf/HashSet.h>
#include <wtf/PassOwnPtr.h>
#include <wtf/HashSet.h>
#include <wtf/Vector.h>
#include <wtf/text/TextPosition.h>

namespace WTF {
class OrdinalNumber;
}

namespace WebCore {

class ContentSecurityPolicyDirective;
class ContentSecurityPolicyDirectiveList;
class ContentSecurityPolicySource;
class ScriptCallStack;
class DOMStringList;
class Frame;
class JSDOMWindowShell;
class ResourceRequest;
class ScriptExecutionContext;
class SecurityOrigin;
class TextEncoding;
class KURL;

typedef Vector<OwnPtr<ContentSecurityPolicyDirectiveList> > CSPDirectiveListVector;
typedef int SandboxFlags;

class ContentSecurityPolicy {
    WTF_MAKE_FAST_ALLOCATED;
public:
    explicit ContentSecurityPolicy(ScriptExecutionContext&);
    explicit ContentSecurityPolicy(const SecurityOrigin&, const Frame* = 0);
    ~ContentSecurityPolicy();

    void copyStateFrom(const ContentSecurityPolicy*);
    void copyUpgradeInsecureRequestStateFrom(const ContentSecurityPolicy*);

    void didCreateWindowShell(JSDOMWindowShell&) const;

    enum class PolicyFrom {
        HTTPEquivMeta,
        HTTPHeader,
        Inherited,
    };
    ContentSecurityPolicyResponseHeaders responseHeaders() const;
    enum ReportParsingErrors { No, Yes };
	void didReceiveHeaders(const ContentSecurityPolicyResponseHeaders&, ReportParsingErrors = ReportParsingErrors::Yes);
	void processHTTPEquiv(const String& content, ContentSecurityPolicyHeaderType type) { didReceiveHeader(content, type, ContentSecurityPolicy::PolicyFrom::HTTPEquivMeta); }

    bool allowScriptWithNonce(const String& nonce, bool overrideContentSecurityPolicy = false) const;
    bool allowStyleWithNonce(const String& nonce, bool overrideContentSecurityPolicy = false) const;

    bool allowJavaScriptURLs(const String& contextURL, const WTF::OrdinalNumber& contextLine, bool overrideContentSecurityPolicy = false) const;
    bool allowInlineEventHandlers(const String& contextURL, const WTF::OrdinalNumber& contextLine, bool overrideContentSecurityPolicy = false) const;
    bool allowInlineScript(const String& contextURL, const WTF::OrdinalNumber& contextLine, const String& scriptContent, bool overrideContentSecurityPolicy = false) const;
    bool allowInlineStyle(const String& contextURL, const WTF::OrdinalNumber& contextLine, const String& styleContent, bool overrideContentSecurityPolicy = false) const;

    bool allowEval(PassRefPtr<ScriptCallStack> callStack, bool overrideContentSecurityPolicy = false) const;

    bool allowPluginType(const String& type, const String& typeAttribute, const KURL&, bool overrideContentSecurityPolicy = false) const;

    bool allowFrameAncestors(const Frame&, const KURL&, bool overrideContentSecurityPolicy = false) const;

    enum class RedirectResponseReceived { No, Yes };
    bool allowScriptFromSource(const KURL&, bool overrideContentSecurityPolicy = false, RedirectResponseReceived = RedirectResponseReceived::No) const;
    bool allowChildFrameFromSource(const KURL&, bool overrideContentSecurityPolicy = false, RedirectResponseReceived = RedirectResponseReceived::No) const;
    bool allowChildContextFromSource(const KURL&, bool overrideContentSecurityPolicy = false, RedirectResponseReceived = RedirectResponseReceived::No) const;
    bool allowImageFromSource(const KURL&, bool overrideContentSecurityPolicy = false, RedirectResponseReceived = RedirectResponseReceived::No) const;
    bool allowStyleFromSource(const KURL&, bool overrideContentSecurityPolicy = false, RedirectResponseReceived = RedirectResponseReceived::No) const;
    bool allowFontFromSource(const KURL&, bool overrideContentSecurityPolicy = false, RedirectResponseReceived = RedirectResponseReceived::No) const;
    bool allowMediaFromSource(const KURL&, bool overrideContentSecurityPolicy = false, RedirectResponseReceived = RedirectResponseReceived::No) const;
    bool allowConnectToSource(const KURL&, bool overrideContentSecurityPolicy = false, RedirectResponseReceived = RedirectResponseReceived::No) const;
    bool allowFormAction(const KURL&, bool overrideContentSecurityPolicy = false, RedirectResponseReceived = RedirectResponseReceived::No) const;

    bool allowObjectFromSource(const KURL&, bool overrideContentSecurityPolicy = false, RedirectResponseReceived = RedirectResponseReceived::No) const;
    bool allowBaseURI(const KURL&, bool overrideContentSecurityPolicy = false) const;

    void setOverrideAllowInlineStyle(bool);

    bool isActive() const;

    void gatherReportURIs(DOMStringList&) const;

    // The following functions are used by internal data structures to call back into this object when parsing, validating,
    // and applying a Content Security Policy.
    // FIXME: We should make the various directives serve only as state stores for the parsed policy and remove these functions.
    // This class should traverse the directives, validating the policy, and applying it to the script execution context.

    // Used by MediaListDirective
    void reportInvalidPluginTypes(const String&) const;

    // Used by CSPSourceList
    void reportDirectiveAsSourceExpression(const String& directiveName, const String& sourceExpression) const;
    void reportInvalidPathCharacter(const String& directiveName, const String& value, const char) const;
    void reportInvalidSourceExpression(const String& directiveName, const String& source) const;
    bool urlMatchesSelf(const KURL&) const;

    // Used by CSPDirectiveList
    void reportDuplicateDirective(const String&) const;
	void reportInvalidDirectiveValueCharacter(const String& directiveName, const String& value) const;
    void reportInvalidSandboxFlags(const String&) const;
    void reportInvalidDirectiveInReportOnlyMode(const String&) const;
    void reportInvalidDirectiveInHTTPEquivMeta(const String&) const;
    void reportMissingReportURI(const String&) const;
    void reportUnsupportedDirective(const String&) const;
    void enforceSandboxFlags(SandboxFlags sandboxFlags) { m_sandboxFlags |= sandboxFlags; }

	// Used by CSPSource
	bool protocolMatchesSelf(const KURL&) const;

    void setUpgradeInsecureRequests(bool);
    bool upgradeInsecureRequests() const { return m_upgradeInsecureRequests; }
    enum class InsecureRequestType { Load, FormSubmission, Navigation };
    void upgradeInsecureRequestIfNeeded(ResourceRequest&, InsecureRequestType) const;
    void upgradeInsecureRequestIfNeeded(KURL&, InsecureRequestType) const;

    void takeNavigationRequestsToUpgrade(HashSet<SecurityOriginData>&);
    void inheritInsecureNavigationRequestsToUpgradeFromOpener(const ContentSecurityPolicy&);
    void setInsecureNavigationRequestsToUpgrade(const HashSet<SecurityOriginData>&);

    void addHashAlgorithmsForInlineScripts(HashSet<ContentSecurityPolicyHashAlgorithm> hashAlgorithmsForInlineScripts)
	{
    	for(HashSet<ContentSecurityPolicyHashAlgorithm>::iterator it = hashAlgorithmsForInlineScripts.begin(); it != hashAlgorithmsForInlineScripts.end(); ++it)
    		m_hashAlgorithmsForInlineScripts.add(*it);
	}
	void addHashAlgorithmsForInlineStylesheets(HashSet<ContentSecurityPolicyHashAlgorithm> hashAlgorithmsForInlineStylesheets)
	{
    	for(HashSet<ContentSecurityPolicyHashAlgorithm>::iterator it = hashAlgorithmsForInlineStylesheets.begin(); it != hashAlgorithmsForInlineStylesheets.end(); ++it)
		m_hashAlgorithmsForInlineStylesheets.add(*it);
	}

private:
    void logToConsole(const String& message, const String& contextURL = String(), const WTF::OrdinalNumber& contextLine = WTF::OrdinalNumber::beforeFirst(), PassRefPtr<ScriptCallStack> = nullptr) const;
    void applyPolicyToScriptExecutionContext();

    void didReceiveHeader(const String&, ContentSecurityPolicyHeaderType, ContentSecurityPolicy::PolicyFrom);

    const TextEncoding& documentEncoding() const;

public:
    enum class Disposition {
        Enforce,
        ReportOnly,
    };

    void reportViolation(const String& violatedDirective, const ContentSecurityPolicyDirective& effectiveViolatedDirective, const KURL& blockedURL, const String& consoleMessage, PassRefPtr<ScriptCallStack>) const;
    void reportViolation(const String& violatedDirective, const ContentSecurityPolicyDirective& effectiveViolatedDirective, const KURL& blockedURL, const String& consoleMessage, const String& sourceURL, const TextPosition& sourcePosition, PassRefPtr<ScriptCallStack> = nullptr) const;
    void reportBlockedScriptExecutionToInspector(const String& directiveText) const;

private:
    // We can never have both a script execution context and a frame.
    ScriptExecutionContext* m_scriptExecutionContext;
    const Frame* m_frame;
    OwnPtr<ContentSecurityPolicySource> m_selfSource;
	String m_selfSourceProtocol;
	CSPDirectiveListVector m_policies;
	String m_lastPolicyEvalDisabledErrorMessage;
	SandboxFlags m_sandboxFlags;
	bool m_overrideInlineStyleAllowed;
	bool m_isReportingEnabled;
	bool m_upgradeInsecureRequests;
    HashSet<ContentSecurityPolicyHashAlgorithm> m_hashAlgorithmsForInlineScripts;
    HashSet<ContentSecurityPolicyHashAlgorithm> m_hashAlgorithmsForInlineStylesheets;
    HashSet<SecurityOriginData> m_insecureNavigationRequestsToUpgrade;
};

}

#endif
